import time
from PIL import Image, ImageSequence
from .tty import output_frame


def _fill_empty(frame):
    canvas = Image.new("RGB", frame.size, (0xFF, 0xFF, 0xFF))
    canvas.paste(frame)
    return canvas


def extract_frames(file_name, width, height, fill=False, **kwargs):
    image = Image.open(file_name)
    for frame in ImageSequence.Iterator(image):
        if fill:
            frame = _fill_empty(image)
        yield frame.resize((width, height), Image.BICUBIC)


def gif_tty(file_name, width, height, sleep=0.02, **kwargs):
    for frame in extract_frames(file_name, width, height, **kwargs):
        output_frame(frame, **kwargs)
        time.sleep(sleep)



import numpy as np
from PIL import ImageFont
from scipy.spatial import cKDTree
from .utils import cache

ASCII_VISIBLE_CHARS = [chr(i) for i in range(32, 127)]


def _char_weight(char):
    font = ImageFont.load_default()
    chr_image = font.getmask(char)
    width, height = font.getsize(char)

    black_pixels = sum(1.0 for i in np.asarray(chr_image) if i > 0)
    return black_pixels / (width * height)


def _normalize_transform(weights):
    weights_range = max(weights) - min(weights)
    return (weights - min(weights)) / weights_range


def _default_weights():
    weights = np.array([_char_weight(c) for c in ASCII_VISIBLE_CHARS])
    return _normalize_transform(weights)


DEFAULT_WEIGHTS = _default_weights()
DEFAULT_WEIGHTS_KD_TREE = cKDTree(DEFAULT_WEIGHTS.reshape(-1, 1))


@cache(key=lambda args: args[0])
def fitting_ascii(rgb_pixel):
    r, g, b = rgb_pixel
    # convert rgb to grayscale
    pixel_weight = (0.2989 * r + 0.5870 * g + 0.1140 * b) / 255
    _, index = DEFAULT_WEIGHTS_KD_TREE.query([pixel_weight])
    return ASCII_VISIBLE_CHARS[index]



import functools


def cache(key):
    def decorator(func):
        _dict = {}

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            cache_key = key(args)

            if cache_key in _dict:
                return _dict[cache_key]

            result = func(*args, **kwargs)
            _dict[cache_key] = result
            return result

        return wrapper

    return decorator



from .gif import gif_tty


def main():
    import argparse

    parser = argparse.ArgumentParser(description="show gif in console mode")
    parser.add_argument("file", help="gif file name")
    parser.add_argument("--width", help="output width", type=int, default=60)
    parser.add_argument("--height", help="output height", type=int, default=30)
    parser.add_argument("--sleep", help="sleep time", type=float, default=0.02)
    parser.add_argument("--fill", help="fill empty", action="store_true")
    parser.add_argument("--color", help="show color", action="store_true")
    args = parser.parse_args()
    gif_tty(
        args.file,
        args.width,
        args.height,
        sleep=args.sleep,
        fill=args.fill,
        color=args.color,
    )


if __name__ == "__main__":
    main()



import sys
from .ascii import fitting_ascii
from .csi import (
    fitting_sgr,
    CUU_CODE,
    SGR_8BIT_FG_COLOR_GRAY,
    SGR_8BIT_FG_COLOR_WHITE,
    SGR_8BIT_BG_COLOR_CODE,
    SGR_8BIT_BG_COLOR_BLACK,
)


def convert_frame(frame, color=True, **kwargs):
    width, height = frame.size

    rgb_frame = frame.convert(mode="RGB")
    rgb_pixels = rgb_frame.load()

    ascii_frame = []
    for y in range(height):
        line = []
        for x in range(width):
            rgb_pixel = rgb_pixels[x, y]
            # convert pixel to ascii and add fg and bg color
            if color:
                line.append(SGR_8BIT_BG_COLOR_CODE % fitting_sgr(rgb_pixel))
                line.append(SGR_8BIT_FG_COLOR_GRAY)
            line.append(fitting_ascii(rgb_pixel))
        # add return for each line
        if color:
            line.append(SGR_8BIT_BG_COLOR_BLACK)
            line.append(SGR_8BIT_FG_COLOR_WHITE)
        line.append("\n")
        ascii_frame.append("".join(line))
    return ascii_frame


def output_frame(frame, backspace=True, **kwargs):
    ascii_frame = convert_frame(frame, **kwargs)
    sys.stdout.writelines(ascii_frame)

    if backspace:
        sys.stdout.write(CUU_CODE % len(ascii_frame))
    sys.stdout.flush()
